// ********************************************************************
// ASTRO-Schaltuhr/Dmmerungsschalter
// AVR-Version / GGC 4.3.2
// LCD DIP204-4 4x20 (bei Bedarf)
// Version 1.0 / Stand: 11.01.2010
// AVR-Version / GGC 4.3.2 / AVR-Studio 4.18
// Mikrocontroller: ATmega32
// Frequenz: 11.059.200 Hz
// ********************************************************************
// FLASH-Speicherbedarf:  ca. 14,2 kByte
// SRAM-Speicherbedarf:   ca. 350 Byte
// EEPROM-Speicherbedarf: 244 Byte
// ********************************************************************
// Da die Sonnenberechnung in Float mit Winkelfunktionen erfolgt, ist
// die Einbindung zustzlicher GCC-Bibliotheken erforderlich:
// Libraries: libprintf_flt.a, libm.a
// Custom Options (Linker Options) -Wl,-u,vfprintf -lprintf_flt 
// ********************************************************************
// Die Programmfunktionen drfen fr private Nutzung verwendet und
// verndert werden. Fr gewerbliche Nutzung ist die Zustimmung
// der Authoren erforderlich.
// Kontakt: Matthias Kahnt, e-mail: az51@gmx.net
// ********************************************************************
// Die Nutzung erfolgt auf eigene Gefahr.
// Fr eventuell entstehende Schden wird keine Haftung bernommen.
// ********************************************************************
//
//************************************************************************
// AVR-Typ: ATmega32
//************************************************************************
//
//************************************************************************
// Includebereich
//************************************************************************

#define F_CPU      11059200UL   // Taktfrequenz AVR
#define SCL_CLOCK     50000UL   // Taktfrequenz I2C-Bus
#define BAUDRATE       9600UL   // UART-Baudrate

#include <avr/io.h>             // IO-Grundfunktionen
#include <avr/interrupt.h>      // Interrupt-Funktionen
#include <avr/pgmspace.h>       // PROGMEM-Funktionen
#include <avr/eeprom.h>         // EEPROM-Funktionen
#include <string.h>             // Stringfunktion
#include <stdio.h>              // Standardfunktion
#include <stdlib.h>             // Standardbibliothek
#include <math.h>               // Mathematische und Winkelfunktionen
#include <util/delay.h>         // Pausenfunktionen
#include <util/twi.h>           // I2C-Funbktionen

//************************************************************************
// Definebereich
//************************************************************************

#define K1_PIN          PD6
#define K2_PIN          PD7
#define LED1_PIN        PA4
#define LED2_PIN        PA5
#define LED3_PIN        PA6
#define LED4_PIN        PA7
#define JP1_PIN         PB0
#define JP2_PIN         PB1
#define SQW_PIN         PD2

#define LED_PORT        PORTA
#define K_PORT          PORTD
#define SQW_PORT        PORTD
#define J_PORT          PORTB

#define LED_PORT_DDR    DDRA
#define K_PORT_DDR      DDRD
#define SQW_PORT_DDR    DDRD
#define J_PORT_DDR      DDRB

#define SQW_PORT_PIN    PIND
#define J_PORT_PIN      PINB

#define ADDRTC          0xD0

//************************************************************************
//************************************************************************
#define LED_D1_EIN    LED_PORT &= ~(1<<LED1_PIN)        /* LED D1 */
#define LED_D1_AUS    LED_PORT |=  (1<<LED1_PIN)

#define LED_D2_EIN    LED_PORT &= ~(1<<LED2_PIN)        /* LED D2 */
#define LED_D2_AUS    LED_PORT |=  (1<<LED2_PIN)

#define LED_D3_EIN    LED_PORT &= ~(1<<LED3_PIN)        /* LED D3 */
#define LED_D3_AUS    LED_PORT |=  (1<<LED3_PIN)

#define LED_D4_EIN    LED_PORT &= ~(1<<LED4_PIN)        /* LED D4 */
#define LED_D4_AUS    LED_PORT |=  (1<<LED4_PIN)

#define K1_EIN        K_PORT |=  (1<<K1_PIN)
#define K1_AUS        K_PORT &= ~(1<<K1_PIN)            /* Relais K1 */

#define K2_EIN        K_PORT |=  (1<<K2_PIN)
#define K2_AUS        K_PORT &= ~(1<<K2_PIN)            /* Relais K2 */

#define JUMPER_JP1    (J_PORT_PIN & (1<<JP1_PIN))       /* Jumper JP1 */
#define JUMPER_JP2    (J_PORT_PIN & (1<<JP2_PIN))       /* Jumper JP2 */

//************************************************************************
// Variablenbereich
//************************************************************************

volatile unsigned char i2c_status     = 0;
volatile unsigned char sys_status     = 0;
volatile unsigned char delay_counter  = 0;
volatile unsigned char uhrzeit_set    = 0;
volatile unsigned char sonne_set      = 0;
volatile unsigned char spr_index      = 0;

// Struktur der Anwender- und Systemeinstellungen (spter im EEPROM)
struct awd_struktur {  
  unsigned char astro_standort;                 // Astro Standort 0 .. 122
  unsigned char astro_daemm;                    // Astro Dmmerungsart 0=brgerlich(-6) 1=nautisch(-12) 2=astronomisch(-18)
  unsigned char software_version;               // 1,2,3 u.s.w.
  unsigned char eeprom_code;                    // 0x55=EEPROM enthlt Systemeinstellungen
};
volatile struct awd_struktur awd;               // Variablen der Einstellungen
struct awd_struktur awd_ee EEMEM;               // dto., nochmal im EEPROM

// Struktur der UART-Daten
#define UART_BUFFER_SIZE 50
struct uart_struktur {  
  unsigned char index;                          // Index der Bufferbelegung
  unsigned char block_laenge;                   // Blocklnge des aktuellen Datenpaketes
  unsigned char block_ok;                       // 1=Block korrekt empfangen
  unsigned char crc_soll;                       // Prfsumme vom PC berechnet
  unsigned char crc_ist;                        // Prfsumme vom uC berechnet
  unsigned char echo;                           // Echofreigabe (fr einige Kommandos abschaltbar)
  unsigned char buffer[UART_BUFFER_SIZE];       // Daten-Buffer
};
volatile struct uart_struktur uart;             // Variablen zur RS-232 Kommunikation

// Zeitdatensatz
struct time {
  unsigned char jahr;                           // Jahr 2 stellen
  unsigned char monat;                          // Monat 1-12
  unsigned char tag;                            // Tag 1-31
  unsigned char stunde;                         // 0-23
  unsigned char minute;                         // 0-59
  unsigned char sekunde;                        // 0-59
  unsigned char wo_tag;                         // 1-7 1=Mo 2=Di ...
  unsigned char mesz;                           // Sommerzeit 1=MESZ(Sommerzeit) 0=MEZ(Winter)
  unsigned char sauf_stunde;                    // Sonnenaufgangszeit 0-23 Stunde
  unsigned char sauf_minute;                    // Sonnenaufgangszeit 0-59 Minute
  unsigned char sunt_stunde;                    // Sonnenuntergangszeit 0-23 Stunde
  unsigned char sunt_minute;                    // Sonnenuntergangszeit 0-59 Minute
  unsigned char md_stunde;                      // Morgendmmerung 0-23 Stunde
  unsigned char md_minute;                      // Morgendmmerung 0-59 Minute
  unsigned char ad_stunde;                      // Abenddmmerung 0-23 Stunde
  unsigned char ad_minute;                      // Abenddmmerung 0-59 Minute
  char          zg_minute;                      // Zeitgleichung Abweichung der Mittagssonne von 12:00Uhr (Vorzeichen +/-) 
};
volatile struct time zeit;                      // aktuelle Uhrzeit 
volatile struct time zeit_t;                    // Uhrzeit vom PC zum Stellen

// Sonnenprogramm-Datensatz
struct sprogramm {
  unsigned char e_fkt;                          // Ein Funktion 0=SA 1=MD 2=SU 3=AD 4=FZ
  unsigned char e_stunde;                       // Ein Offset Stunde
  unsigned char e_minute;                       // Ein Offset Minute
  unsigned char e_vorz;                         // Ein Vorzeichen 0=positiv 1=negativ 2=bei FZ
  unsigned char wo_tag;                         // Bit7=xx Bit6=Sa Bit5=Fr Bit4=Do Bit3=Mi Bit2=Di Bit1=Mo Bit0=So
  unsigned char a_fkt;                          // Aus Funktion 0=SA 1=MD 2=SU 3=AD 4=FZ
  unsigned char a_stunde;                       // Aus Offset Stunde
  unsigned char a_minute;                       // Aus Offset Minute
  unsigned char a_vorz;                         // Aus Vorzeichen 0=positiv 1=negativ 2=bei FZ
  unsigned char k1;                             // 1=Relais K1
  unsigned char k2;                             // 1=Relais K2
  unsigned char belegt;                         // 0=Eintrag frei 1=Eintrag belegt
};
#define SPR_MAX 20
volatile struct sprogramm spr[SPR_MAX];         // Variablen des Sonnenprogrammes
struct sprogramm spr_ee[SPR_MAX] EEMEM;         // Variablen des Sonnenprogrammes (EEPROM)

// Ausschaltzeiten
struct austime {
  unsigned char stunde;                         // 0-23 Stunde der Auschaltzeit des Relais Kx
  unsigned char minute;                         // 0-59 Minute der Auschaltzeit des Relais Kx
  unsigned char sekunde;                        // 0-59 Sekunde der Auschaltzeit des Relais Kx
  unsigned char zyklus;                         // 1=Schaltaktivitt gerade aktiv
  unsigned char zustand;                        // 1=gerade Ein 0=gerdae Aus
};
volatile struct austime k1aus;                  // Variablen fr Relais K1
volatile struct austime k2aus;                  // Variablen fr Relais K2

// Struktur der geografischen Koordinaten
struct ortskoordinaten_struktur {
  float laenge;                                 // Lngengrad
  float breite;                                 // Breitengrad
};

// Tabelle der Orte mit den geografischen Koordinaten im FLASH
struct ortskoordinaten_struktur ortskoordinaten[] PROGMEM = {
  { /*  00  */   14.9, 51.1 }, /* "Grlitz"   */
  { /*  01  */   13.8, 51.1 }, /* "Dresden"   */
  { /*  02  */   14.4, 51.2 }, /* "Bautzen"   */
  { /*  03  */   14.3, 51.7 }, /* "Cottbus"   */
  { /*  04  */   12.4, 51.3 }, /* "Leipzig"   */
  { /*  05  */   11.3, 51.5 }, /* "Sangerh."  */
  { /*  06  */   11.9, 51.5 }, /* "Halle"     */
  { /*  07  */   12.1, 50.8 }, /* "Gera"      */
  { /*  08  */   12.5, 50.7 }, /* "Zwickau"   */
  { /*  09  */   12.9, 50.8 }, /* "Chemnitz"  */
  { /*  10  */   13.4, 52.5 }, /* "Berlin"    */
  { /*  11  */   12.3, 52.9 }, /* "Kyritz"    */
  { /*  12  */   13.8, 53.3 }, /* "Prenzlau"  */
  { /*  13  */   13.7, 51.8 }, /* "Luckau"    */
  { /*  14  */   12.9, 52.4 }, /* "Potsdam"   */
  { /*  15  */   14.2, 52.3 }, /* "Frkf.Oder" */
  { /*  16  */   13.1, 52.7 }, /* "Oranienb." */
  { /*  17  */   13.3, 53.5 }, /* "Neubrand." */
  { /*  18  */   12.0, 54.1 }, /* "Rostock"   */
  { /*  19  */   11.3, 53.6 }, /* "Schwerin"  */
  { /*  20  */    9.9, 53.5 }, /* "Hamburg"   */
  { /*  21  */    9.4, 53.7 }, /* "Stade"     */
  { /*  22  */   10.0, 53.7 }, /* "Norderst." */
  { /*  23  */   10.9, 53.9 }, /* "Lbeck"    */
  { /*  24  */   10.1, 54.3 }, /* "Kiel"      */
  { /*  25  */    9.1, 54.2 }, /* "Heide"     */
  { /*  26  */    7.1, 53.3 }, /* "Emden"     */
  { /*  27  */    8.5, 53.5 }, /* "Bremerha." */
  { /*  28  */    8.8, 53.1 }, /* "Bremen"    */
  { /*  29  */   10.2, 52.8 }, /* "Celle"     */
  { /*  30  */    9.6, 52.3 }, /* "Hannover"  */
  { /*  31  */    9.8, 52.1 }, /* "Hildesh."  */
  { /*  32  */    8.7, 52.1 }, /* "Herford"   */
  { /*  33  */    8.5, 52.0 }, /* "Bielefeld" */
  { /*  34  */    9.4, 51.3 }, /* "Kassel"    */
  { /*  35  */    8.7, 50.6 }, /* "Giessen"   */
  { /*  36  */    9.7, 50.5 }, /* "Fulda"     */
  { /*  37  */    9.9, 51.5 }, /* "Gttingen" */
  { /*  38  */   10.5, 52.3 }, /* "Brauns."   */
  { /*  39  */   11.6, 52.2 }, /* "Magdeburg" */
  { /*  40  */    6.8, 51.2 }, /* "Dsseld."  */
  { /*  41  */    6.4, 51.2 }, /* "Gladbach"  */
  { /*  42  */    7.1, 51.2 }, /* "Wuppertal" */
  { /*  43  */    6.6, 51.8 }, /* "Bocholt"   */
  { /*  44  */    7.5, 51.5 }, /* "Dortmund"  */
  { /*  45  */    7.0, 51.4 }, /* "Essen"     */
  { /*  46  */    6.9, 51.5 }, /* "Bottrop"   */
  { /*  47  */    6.7, 51.4 }, /* "Duisburg"  */
  { /*  48  */    7.6, 51.9 }, /* "Mnster"   */
  { /*  49  */    8.0, 51.3 }, /* "Osnabrck" */
  { /*  50  */    6.9, 51.0 }, /* "Kln"      */
  { /*  51  */    7.5, 50.9 }, /* "Wiehl"     */
  { /*  52  */    6.1, 50.8 }, /* "Aachen"    */
  { /*  53  */    7.1, 50.7 }, /* "Bonn"      */
  { /*  54  */    6.6, 49.7 }, /* "Trier"     */
  { /*  55  */    8.0, 49.9 }, /* "Mainz"     */
  { /*  56  */    7.6, 50.3 }, /* "Koblenz"   */
  { /*  57  */    8.0, 50.8 }, /* "Siegen"    */
  { /*  58  */    7.5, 51.3 }, /* "Hagen"     */
  { /*  59  */    7.8, 51.6 }, /* "Hamm"      */
  { /*  60  */    8.7, 50.1 }, /* "Frankfurt" */
  { /*  61  */    8.1, 50.4 }, /* "Limburg"   */
  { /*  62  */    8.2, 50.7 }, /* "Herborn"   */
  { /*  63  */    9.2, 49.9 }, /* "Aschaffb." */
  { /*  64  */    8.6, 49.8 }, /* "Darmstadt" */
  { /*  65  */    8.1, 50.0 }, /* "Wiesbaden" */
  { /*  66  */    7.0, 49.2 }, /* "Saarbrck."*/
  { /*  67  */    7.7, 49.4 }, /* "Kaisersl." */
  { /*  68  */    8.4, 49.4 }, /* "Mannheim"  */
  { /*  69  */    8.6, 49.4 }, /* "Heidelb."  */
  { /*  70  */    9.2, 48.7 }, /* "Stuttgart" */
  { /*  71  */    9.2, 58.9 }, /* "Ludwigsb." */
  { /*  72  */    9.1, 48.5 }, /* "Tbingen"  */
  { /*  73  */    9.7, 48.7 }, /* "Gppingen" */
  { /*  74  */    9.4, 49.3 }, /* "Heilbronn" */
  { /*  75  */    8.7, 48.9 }, /* "Pforzheim" */
  { /*  76  */    8.4, 49.0 }, /* "Karlsruhe" */
  { /*  77  */    7.9, 48.4 }, /* "Offenburg" */
  { /*  78  */    9.1, 47.6 }, /* "Konstanz"  */
  { /*  79  */    7.8, 48.0 }, /* "Freiburg"  */
  { /*  80  */   11.5, 48.2 }, /* "Mnchen"   */
  { /*  81  */   10.8, 48.0 }, /* "Landsberg" */
  { /*  82  */   11.1, 47.5 }, /* "Garmisch"  */
  { /*  83  */   12.2, 47.8 }, /* "Rosenheim" */
  { /*  84  */   12.2, 48.5 }, /* "Landshut"  */
  { /*  85  */   11.4, 48.7 }, /* "Ingolsta." */
  { /*  86  */   10.9, 48.4 }, /* "Augsburg"  */
  { /*  87  */   10.3, 47.7 }, /* "Kempten"   */
  { /*  88  */    9.6, 47.6 }, /* "Tettnang"  */
  { /*  89  */   10.0, 48.3 }, /* "Ulm"       */
  { /*  90  */   11.1, 49.4 }, /* "Nrnberg"  */
  { /*  91  */   10.6, 49.3 }, /* "Anspach"   */
  { /*  92  */   11.9, 49.4 }, /* "Amberg"    */
  { /*  93  */   12.1, 49.0 }, /* "Regensbg." */
  { /*  94  */   13.4, 48.6 }, /* "Passau"    */
  { /*  95  */   11.9, 50.2 }, /* "Hof"       */
  { /*  96  */   10.9, 49.9 }, /* "Bamberg"   */
  { /*  97  */    9.9, 49.7 }, /* "Wrzburg"  */
  { /*  98  */   10.6, 50.6 }, /* "Suhl"      */
  { /*  99  */   11.0, 50.1 }, /* "Erfurt"    */
  { /* 100  */   13.3, 54.4 }, /* "Rgen"     */
  { /* 101  */   13.8, 54.1 }, /* "Usedom"    */
  { /* 102  */   11.2, 54.4 }, /* "Fehmann"   */
  { /* 103  */    8.3, 54.9 }, /* "Sylt"      */
  { /* 104  */    7.8, 54.1 }, /* "Helgoland" */
  { /* 105  */    7.5, 53.7 }, /* "Langeoog"  */
  { /* 106  */   16.4, 48.1 }, /* "Wien"    A */
  { /* 107  */   14.2, 48.3 }, /* "Linz"      */
  { /* 108  */   15.4, 47.0 }, /* "Graz"      */
  { /* 109  */   14.3, 46.6 }, /* "Klagenfu." */
  { /* 110  */   11.4, 47.2 }, /* "Innsbruck" */
  { /* 111  */    9.8, 47.2 }, /* "Bludenz"   */
  { /* 112  */   15.6, 48.6 }, /* "Horn"      */
  { /* 113  */    8.5, 47.3 }, /* "Zrich" CH */
  { /* 114  */    7.5, 46.8 }, /* "Bern"      */
  { /* 115  */    8.9, 45.9 }, /* "Lugano"    */
  { /* 116  */    6.2, 46.2 }, /* "Genf"      */
  { /* 117  */   14.4, 50.1 }, /* "Praha"  CZ */
  { /* 118  */   13.4, 49.7 }, /* "Pilsen"    */
  { /* 119  */   14.6, 49.3 }, /* "Tabor"     */
  { /* 120  */   16.6, 49.2 }, /* "Brno"      */
  { /* 121  */   18.3, 49.8 }, /* "Ostrava"   */
  { /* 122  */   15.1, 50.8 }  /* "Liberec"   */
}; 

#include "ds1307_avr_twi.c"     // DS1307 RTC Basis-Funktionen
#include "usart_avr.c"          // UART-Funktionen

// ********************************************************************
// System-Initialisierungen
// ********************************************************************
void init_system (void) {
  // Portdefinition
  LED_PORT_DDR |= ((1 << LED1_PIN) | (1 << LED2_PIN) | (1 << LED3_PIN) | (1 << LED4_PIN)) ;   // LED D1,D2,D3,D4 (Ausgang)
  K_PORT_DDR   |= ((1 << K1_PIN)   | (1 << K2_PIN));    // K1,K2 (Ausgnge)
  SQW_PORT_DDR &= ~(1 << SQW_PIN);                      // SWQ (Eingang)
  SQW_PORT     |=    (1<<SQW_PIN);                      // Pull-Up-R zuschalten
  J_PORT_DDR   &= ~((1 << JP1_PIN) | (1 << JP2_PIN));   // JP1,JP2 (Eingang)
  J_PORT       |=  ((1 << JP1_PIN) | (1 << JP2_PIN));   // Pull-Up-R zuschalten

  LED_D1_AUS;                       // alle LEDs aus
  LED_D2_AUS;  
  LED_D3_AUS;
  LED_D4_AUS;
  K1_AUS;                           // beide Relais aus
  K2_AUS;

  i2c_init();                       // RTC initialisieren
  usart_init(BAUDRATE,F_CPU);       // UART initialisieren

  // RTC auf Vorhandensein testen
  i2c_status = i2c_start(ADDRTC);
  if (i2c_status) {
    // Fehler
    LED_D1_EIN;
    sys_status = 0x01;              // I2C-Fehler
  } 
  i2c_stop();
  
  k1aus.stunde    = 0;              // K1 Ausschaltzeiten initialisieren
  k1aus.minute    = 0;
  k1aus.sekunde   = 0;
  k1aus.zyklus    = 0;
  k1aus.zustand   = 0;

  k2aus.stunde    = 0;              // K2 Ausschaltzeiten initialisieren
  k2aus.minute    = 0;
  k2aus.sekunde   = 0;
  k2aus.zyklus    = 0;
  k2aus.zustand   = 0;

  sonne_set       = 1;              // Sonnenberechnung anstossen

  // Parameterdaten aus EEPROM laden
  eeprom_read_block((void*)&awd,&awd_ee,sizeof(awd));

  // Astroprogramm aus EEPROM laden
  eeprom_read_block((void*)&spr,&spr_ee,sizeof(spr));

  // Prfen, ob ein Astro-Schaltprogramm im Speicher ist und ggf. LED einschalten
  for(spr_index = 0; spr_index < SPR_MAX; spr_index++) {
    if (spr[spr_index].belegt) LED_D2_EIN;  // LED D2 ein
  }

  // Prfen auf Werkseinstellungen
  if ((!JUMPER_JP1) || (awd.eeprom_code != 0x55)) {
    // Grundinitialisierungen bei Erstinbetriebnahme oder Werkseinstellungen
    awd.astro_standort          = 34;   // Standort
    awd.astro_daemm             = 0;    // Astro Dmmerungsart 0=brgerlich(-6) 1=nautisch(-12) 2=astronomisch(-18)
    awd.software_version        = 1;    // Version
    awd.eeprom_code             = 0x55; // EEPROM-Kennbyte

    // Astroprogramm lschen
    for(spr_index = 0; spr_index < SPR_MAX; spr_index++) {
        spr[spr_index].e_fkt    = 0;
        spr[spr_index].e_stunde = 0;
        spr[spr_index].e_minute = 0;
        spr[spr_index].e_vorz   = 0;
        spr[spr_index].wo_tag   = 0;
        spr[spr_index].a_fkt    = 0;
        spr[spr_index].a_stunde = 0;
        spr[spr_index].a_minute = 0;
        spr[spr_index].a_vorz   = 0;
        spr[spr_index].k1       = 0;
        spr[spr_index].k2       = 0;
        spr[spr_index].belegt   = 0;
    } 

    LED_D2_AUS;  // LED D2 aus

    // Parameterdaten in EEPROM speichern
    eeprom_write_block((void*)&awd,&awd_ee,sizeof(awd));

    // Astroprogramm in EEPROM speichern
    eeprom_write_block((void*)&spr,&spr_ee,sizeof(spr));

    // Zeitdatensatz Starteinstellungen
    zeit.jahr           = 10;           // 01.01.2010, Freitag
    zeit.monat          = 1;
    zeit.tag            = 1;
    zeit.stunde         = 0;
    zeit.minute         = 0;
    zeit.sekunde        = 0;
    zeit.wo_tag         = 5;            // 1=Mo. 2=Die. 3=Mi. 4=Do. 5=Fr. 6=Sa. 7=So.
    zeit.mesz           = 0;
    zeit.sauf_stunde    = 0;
    zeit.sauf_minute    = 0;
    zeit.sunt_stunde    = 0;
    zeit.sunt_minute    = 0;
    zeit.md_stunde      = 0;
    zeit.md_minute      = 0;
    zeit.ad_stunde      = 0;
    zeit.ad_minute      = 0;
    zeit.zg_minute      = 0;     
    
    // RTC Anfangswerte setzen
    i2c_start_wait(ADDRTC);             // Start (Write)
    i2c_write(0x00);
    i2c_write(bin_bcd(zeit.sekunde));   // Sekunden Oszi=Start
    i2c_write(bin_bcd(zeit.minute));    // Minuten
    i2c_write(bin_bcd(zeit.stunde));    // Stunden
    i2c_write(zeit.wo_tag);             // Wochentag
    i2c_write(bin_bcd(zeit.tag));       // Tag
    i2c_write(bin_bcd(zeit.monat));     // Monat
    i2c_write(bin_bcd(zeit.jahr));      // Jahr
    i2c_write(0x10);                    // Control: 0001 0000 SQWE=1 RS1=0 RS0=0 --> 1Hz Signal am SQW/OUT-PIN
    i2c_write(zeit.mesz);               // Zeitzone im NOV-RAM Adresse 0x08 speichern
    i2c_stop();         
  }

  // Externer Interrupt 0
  GICR  |=  (1 << INT0);                // INT0 Interrupt freigeben
  MCUCR |=  (1 << ISC01);               // H/L-Flanke lst Interrupt aus (11: L/H, 10: H/L)

  // Timer 0 Interrupt
  TCCR0 = 0x05;                         // Timer 0 1024-Vorteiler (92.6us/Takt bei 11.0592MHz)
  TIMSK = 0x01;                         // Timer 0 Overflow  Interrupt freigeben

  // Generelle Interruptfreigabe
  sei();
}

// ********************************************************************
// Hauptprogramm
// ********************************************************************
int main (void) {
  init_system();                        // System Init
  while(1) {                            // Hauptschleife
    // UART-Empfang vorbereiten
    for (uart.index = 0; uart.index < UART_BUFFER_SIZE; uart.index++) uart.buffer[uart.index] = 0; // Buffer lschen 
    uart.echo         = 1;
    uart.crc_soll     = 0;
    uart.crc_ist      = 0;
    uart.block_laenge = UART_BUFFER_SIZE - 1;
    uart.index        = 1;
    uart.block_ok     = 0;
    // auf UART-Empfang warten
    while(!uart.block_ok);
    // Datenblock per UART liegt vor
    LED_D1_EIN;
    switch (uart.buffer[1]) {
            case 'V':   // verbinden
                        usart_send_byte('V');   // Echo (immer)
                        break;
            case 'C':   // trennen
                        usart_send_byte('C');   // Echo (immer)
                        break;
            case 'T':   // testen
                        usart_send_byte('T');   // Echo (immer)
                        break;
            case 'E':   // Astro-Schaltprogramm in EEPROM speichern
                        usart_send_byte('E');   // Echo (immer)
                        eeprom_write_block((void*)&spr,&spr_ee,sizeof(spr));
                        break;
            case 'R':   // Einstellungen (Reverse) zurck zum PC senden
                        uart.buffer[1]  = 'R';                    // Kennbyte
                        uart.buffer[2]  = 31;                     // Anzahl der Bytes
                        uart.buffer[3]  = 0;                      // keine Bedeutung
                        
                        uart.buffer[4]  = zeit.tag;               // Datum
                        uart.buffer[5]  = zeit.monat;
                        uart.buffer[6]  = zeit.jahr;
                        uart.buffer[7]  = zeit.wo_tag;
                        uart.buffer[8]  = zeit.stunde;            // Uhrzeit
                        uart.buffer[9]  = zeit.minute;
                        uart.buffer[10] = zeit.sekunde;
                        uart.buffer[11] = zeit.mesz;

                        uart.buffer[12] = zeit.md_stunde;        // Morgendmmerung
                        uart.buffer[13] = zeit.md_minute;
                        uart.buffer[14] = zeit.ad_stunde;        // Abenddmmerung
                        uart.buffer[15] = zeit.ad_minute;
                        uart.buffer[16] = zeit.zg_minute;        // Mittagsonne

                        uart.buffer[17] = zeit.sauf_stunde;      // Sonnenaufgang
                        uart.buffer[18] = zeit.sauf_minute;
                        uart.buffer[19] = zeit.sunt_stunde;      // Sonnenuntergang
                        uart.buffer[20] = zeit.sunt_minute;

                        uart.buffer[21] = awd.astro_standort;    // Standort
                        uart.buffer[22] = (unsigned char)   pgm_read_float(&ortskoordinaten[awd.astro_standort].laenge);  // geografische Lnge
                        uart.buffer[23] = (unsigned char) ((pgm_read_float(&ortskoordinaten[awd.astro_standort].laenge) - (int) (pgm_read_float(&ortskoordinaten[awd.astro_standort].laenge))) * 10 + 0.5);
                        uart.buffer[24] = (unsigned char)   pgm_read_float(&ortskoordinaten[awd.astro_standort].breite);  // geografische Breite
                        uart.buffer[25] = (unsigned char) ((pgm_read_float(&ortskoordinaten[awd.astro_standort].breite) - (int) (pgm_read_float(&ortskoordinaten[awd.astro_standort].breite))) * 10 + 0.5);

                        uart.buffer[26] = awd.astro_daemm;       // Astro Dmmerungsart

                        uart.buffer[27] = k1aus.zustand;         // K1 Ist-Zustand 1=Ein 0=Aus
                        uart.buffer[28] = k2aus.zustand;         // K2 Ist-Zustand 1=Ein 0=Aus

                        uart.buffer[29] = awd.software_version;  // Softwareversion
                        uart.buffer[30] = sys_status;            // Fehlerstatus

                        // Prfsumme
                        uart.crc_ist = 0;
                        for (uart.index = 1; uart.index < uart.buffer[2]; uart.index++) uart.crc_ist += uart.buffer[uart.index];
                        uart.buffer[31] = uart.crc_ist;
                        
                        // Buffer senden
                        for (uart.index = 1; uart.index <= uart.buffer[2]; uart.index++) {
                          usart_send_byte(uart.buffer[uart.index]);
                        }
                        break;
            case 'D':   // Astroprogramm (Reverse) zurck zum PC senden
                        for (uart.index = 1; uart.index < uart.block_laenge; uart.index++) uart.crc_ist += uart.buffer[uart.index];
                        uart.index = uart.buffer[3] - 1;            // angeforderte Eintrags-Nr.
                        if ((uart.crc_ist == uart.crc_soll) && (uart.index < 20)) {
                            // Prfsumme stimmt
                            uart.buffer[1]  = 'D';               // Kennbyte
                            uart.buffer[2]  = 16;                // Anzahl der Bytes

                            uart.buffer[4]  = spr[uart.index].wo_tag;
                            uart.buffer[5]  = spr[uart.index].e_fkt;
                            uart.buffer[6]  = spr[uart.index].e_stunde;
                            uart.buffer[7]  = spr[uart.index].e_minute;
                            uart.buffer[8]  = spr[uart.index].e_vorz;
                            uart.buffer[9]  = spr[uart.index].a_fkt;
                            uart.buffer[10] = spr[uart.index].a_stunde;
                            uart.buffer[11] = spr[uart.index].a_minute;
                            uart.buffer[12] = spr[uart.index].a_vorz;
                            uart.buffer[13] = spr[uart.index].k1;
                            uart.buffer[14] = spr[uart.index].k2;
                            uart.buffer[15] = spr[uart.index].belegt;

                            // Prfsumme
                            uart.crc_ist = 0;
                            for (uart.index = 1; uart.index < uart.buffer[2]; uart.index++) uart.crc_ist += uart.buffer[uart.index];
                            uart.buffer[16] = uart.crc_ist;
                        
                            // Buffer senden
                            for (uart.index = 1; uart.index <= uart.buffer[2]; uart.index++) {
                              usart_send_byte(uart.buffer[uart.index]);
                            }
                        }
                        break;
            case 'Z':   // Uhrzeit stellen
                        for (uart.index = 1; uart.index < uart.block_laenge; uart.index++) uart.crc_ist += uart.buffer[uart.index];
                        if (uart.crc_ist == uart.crc_soll) {
                            // Prfsumme stimmt, bernahme vorbereiten
                            zeit_t.tag     = uart.buffer[4];
                            zeit_t.monat   = uart.buffer[5];
                            zeit_t.jahr    = uart.buffer[6];
                            zeit_t.wo_tag  = uart.buffer[7];
                            zeit_t.stunde  = uart.buffer[8];
                            zeit_t.minute  = uart.buffer[9];
                            zeit_t.sekunde = uart.buffer[10];
                            zeit_t.mesz    = uart.buffer[11];
                            uhrzeit_set    = 1;                     // bernahme vorbereiten
                            sonne_set      = 1;                     // Sonnenberechnung anstossen
                            if (uart.echo) usart_send_byte('Z');    // Echo (bei Bedarf)
                        }
                        break;
            case 'M':   // Manuellbetrieb der Relais
                        for (uart.index = 1; uart.index < uart.block_laenge; uart.index++) uart.crc_ist += uart.buffer[uart.index]; // CRC 
                        if (uart.crc_ist == uart.crc_soll) {
                            switch (uart.buffer[4]) {
                                case 0: // Relais K1 aus
                                        K1_AUS;
                                        LED_D3_AUS;
                                        k1aus.zustand = 0;
                                        break;
                                case 1: // Relais K1 ein
                                        K1_EIN;
                                        LED_D3_EIN;
                                        k1aus.zustand = 1;
                                        break;
                            }
                            switch (uart.buffer[5]) {
                                case 0: // Relais K2 aus
                                        K2_AUS;
                                        LED_D4_AUS;
                                        k2aus.zustand = 0;
                                        break;
                                case 1: // Relais K2 ein
                                        K2_EIN;
                                        LED_D4_EIN;
                                        k2aus.zustand = 1;
                                        break;
                            }
                            if (uart.echo) usart_send_byte('M');    // Echo (bei Bedarf)
                        }
                        break;
            case 'P':   // Parameter einstellen
                        for (uart.index = 1; uart.index < uart.block_laenge; uart.index++) uart.crc_ist += uart.buffer[uart.index]; // CRC 
                        if (uart.crc_ist == uart.crc_soll) {
                            // Prfsumme stimmt, Daten bernehmen
                            awd.astro_daemm    = uart.buffer[4];
                            awd.astro_standort = uart.buffer[5];
                            // Parameterdaten in EEPROM speichern
                            eeprom_write_block((void*)&awd,&awd_ee,sizeof(awd));
                            sonne_set = 1;                          // Sonnenberechnung anstossen
                            if (uart.echo) usart_send_byte('P');    // Echo (bei Bedarf)
                        }
                        break;
            case 'S':   // Astro-Schaltprogramm
                        for (uart.index = 1; uart.index < uart.block_laenge; uart.index++) uart.crc_ist += uart.buffer[uart.index]; // CRC 
                        if (uart.crc_ist == uart.crc_soll) {
                            // Prfsumme stimmt
                            spr_index = uart.buffer[3]-1;       // Eintrag-Nr.
                            if (spr_index < SPR_MAX) {
                                // Astroprogramm-Eintrag bernehmen
                                spr[spr_index].wo_tag   = uart.buffer[4];
                                spr[spr_index].e_fkt    = uart.buffer[5];
                                spr[spr_index].e_stunde = uart.buffer[6];
                                spr[spr_index].e_minute = uart.buffer[7];
                                spr[spr_index].e_vorz   = uart.buffer[8];
                                spr[spr_index].a_fkt    = uart.buffer[9];
                                spr[spr_index].a_stunde = uart.buffer[10];
                                spr[spr_index].a_minute = uart.buffer[11];
                                spr[spr_index].a_vorz   = uart.buffer[12];
                                spr[spr_index].k1       = uart.buffer[13];
                                spr[spr_index].k2       = uart.buffer[14];
                                spr[spr_index].belegt   = uart.buffer[15];
                                k1aus.zyklus = 0;               // Falls gerade Schaltzeiten aktiv, rcksetzen
                                k2aus.zyklus = 0;
                                K1_AUS;                         // Relais aus
                                K2_AUS;
                                LED_D2_AUS;                     // LED D2 aus
                                // Prfen, ob ein Astro-Schaltprogramm im Speicher ist und ggf. LED einschalten
                                for(spr_index = 0; spr_index < SPR_MAX; spr_index++) {
                                    if (spr[spr_index].belegt) LED_D2_EIN;  // LED D2 ein
                                }
                            }
                            if (uart.echo) usart_send_byte('S');    // Echo (bei Bedarf)
                        }
                        break;
    } // Switch
    LED_D1_AUS;
  }
}

// ********************************************************************
// ********************************************************************
// INT0 Interruptroutine
// ********************************************************************
ISR(INT0_vect) {
// Variablen, die nur im INT0 verwendet werden
unsigned int jahrestag      = 0;
unsigned char ereignis_flag = 0;
unsigned char s_index       = 0;
int zeit_index              = 0;
int zeit_stopp_index        = 0;
int zeit_start_index        = 0; 
float breite_gr,geogr_l,geogr_b,dk,zd,zg,sa,su,h_const;

  // prfen, ob neue Uhrzeit gestellt werden soll
  if (uhrzeit_set) {
        // neue Uhrzeit einstellen
        i2c_start_wait(ADDRTC);                             // Start (Write)
        i2c_write(0x00);                                    // Adresse 0x00 setzen
        i2c_write(bin_bcd(zeit_t.sekunde));                 // Sekunden Oszi=Start
        i2c_write(bin_bcd(zeit_t.minute));                  // Minuten
        i2c_write(bin_bcd(zeit_t.stunde));                  // Stunden
        i2c_write(zeit_t.wo_tag);                           // Wochentag
        i2c_write(bin_bcd(zeit_t.tag));                     // Tag
        i2c_write(bin_bcd(zeit_t.monat));                   // Monat
        i2c_write(bin_bcd(zeit_t.jahr));                    // Jahr;
        i2c_write(0x10);                                    // Control-Register (1Hz Output)
        i2c_write(zeit_t.mesz);                             // Zeitzone auf NOV-RAM Adresse 0x08
        i2c_stop();                                         // Stop
        uhrzeit_set = 0;                                    // Uhrzeit wurde bernommen
  }

  // Uhrzeit vom RTC bernehmen
  i2c_start_wait(ADDRTC);                                   // Start (Write)
  i2c_write(0x00);                                          // Adresse 0x00 setzen
  i2c_rep_start(ADDRTC + 0x01);                             // Start (Read)
  zeit.sekunde = bcd_bin(i2c_readAck());                    // Datensatz einlesen
  zeit.minute  = bcd_bin(i2c_readAck());
  zeit.stunde  = bcd_bin(i2c_readAck());
  zeit.wo_tag  = bcd_bin(i2c_readAck());
  zeit.tag     = bcd_bin(i2c_readAck());
  zeit.monat   = bcd_bin(i2c_readAck());
  zeit.jahr    = bcd_bin(i2c_readAck());
  zeit.mesz    = bcd_bin(i2c_readAck());                    // Dummy-Read Control-Register
  zeit.mesz    = bcd_bin(i2c_readNak());                    // NOV-RAM Adresse 0x08
  i2c_stop();                                               // Stop

  // MESZ/MEZ Umstellung prfen
  if (zeit.wo_tag == 7) {
        // Sonntag
        if (zeit.monat == 3) {
            // Monat Mrz (Umstellung von MEZ auf MESZ)
            if ((!zeit.mesz) && (zeit.tag > 24)) {
                // Es ist noch Normalzeit und der letzte Sonntag im Monat Mrz
                if ((zeit.stunde == 2) && (!zeit.minute) && (!zeit.sekunde)) {
                    // 2 Uhr Nachts
                    zeit.stunde++;                          // jetzt 1h vorstellen
                    zeit.mesz = 1;                          // MESZ setzen
                    // neue Stunde auch beim RTC setzen
                    i2c_start_wait(ADDRTC);                 // Start (Write)
                    i2c_write(0x02);                        // Adresse 0x02 setzen (Stunde)
                    i2c_write(bin_bcd(zeit.stunde));        // Stunden
                    i2c_stop();                             // Stop
                    i2c_start_wait(ADDRTC);                 // Start (Write)
                    i2c_write(0x08);                        // Adresse NOV-RAM 0x08 setzen
                    i2c_write(zeit.mesz);                   // Zeitzone speichern
                    i2c_stop();                             // Stop
                }
            }
        }
        if (zeit.monat == 10) {
            // Monat Oktober (Umstellung von MESZ auf MEZ)
            if ((zeit.mesz) && (zeit.tag > 24)) {
                // Es ist noch Normalzeit und der letzte Sonntag im Monat Oktober
                if ((zeit.stunde == 3) && (!zeit.minute) && (!zeit.sekunde)) {
                    // 3 Uhr Nachts
                    zeit.stunde--;                          // jetzt 1h zurckstellen
                    zeit.mesz = 0;                          // MEZ setzen
                    // neue Stunde auch beim RTC setzen
                    i2c_start_wait(ADDRTC);                 // Start (Write)
                    i2c_write(0x02);                        // Adresse 0x02 setzen (Stunde)
                    i2c_write(bin_bcd(zeit.stunde));        // Stunden
                    i2c_stop();                             // Stop
                    i2c_start_wait(ADDRTC);                 // Start (Write)
                    i2c_write(0x08);                        // Adresse NOV-RAM 0x08 setzen
                    i2c_write(zeit.mesz);                   // Zeitzone speichern
                    i2c_stop();                             // Stop
                }
            }
        }
  }

  // Sonnenberechung anstossen
  if ((!zeit.minute) && (!zeit.sekunde)) sonne_set = 1;     // Volle Stunde xx:00:00

  // Sonnenberechnungen
  if (sonne_set) {
    #define PI_KONST 3.1415926536
    #define H_KONST_1  (-0.83 *  PI_KONST / 180.0)
    #define H_KONST_6  (-6.0  *  PI_KONST / 180.0)
    #define H_KONST_12 (-12.0 *  PI_KONST / 180.0)
    #define H_KONST_18 (-18.0 *  PI_KONST / 180.0)
    #define ZEITZONE 1

    // Koordinaten aus FLASH holen
    geogr_l = pgm_read_float(&ortskoordinaten[awd.astro_standort].laenge);
    geogr_b = pgm_read_float(&ortskoordinaten[awd.astro_standort].breite);
  
    // Tag des Jahres bestimmen
    jahrestag = 0;
    switch(zeit.monat - 1) {
        case 11: jahrestag += 30;
        case 10: jahrestag += 31;
        case 9:  jahrestag += 30;
        case 8:  jahrestag += 31;
        case 7:  jahrestag += 31;
        case 6:  jahrestag += 30;
        case 5:  jahrestag += 31;
        case 4:  jahrestag += 30;
        case 3:  jahrestag += 31;
        case 2:  jahrestag += 28;
        case 1:  jahrestag += 31;
    }
    if (!(zeit.jahr % 4)) jahrestag += 1;                   // Schaltjahr
    jahrestag += zeit.tag;

    // Sonnendeklination und Zeitgleichung
    breite_gr = geogr_b * PI_KONST / 180.0;                 // Breite umrechnen      
    dk = 0.40954 * sin(0.0172 * (jahrestag - 79.35));       // Sonnendeklination
    zg = -0.1752 * sin(0.033430 * jahrestag + 0.5474) - 0.1340 * sin(0.018234 * jahrestag - 0.1939);    // Zeitgleichung

    // Sonnenaufgangszeit und Sonnenuntergangszeit berechnen
    h_const = H_KONST_1;
    zd = (sin(h_const) - sin(breite_gr) * sin(dk)) / (cos(breite_gr) * cos(dk));
    if (zd > 1.0)    zd =  1.0;                             // Begrenzung, da ARCCOS ist nur von -1 bis +1 definiert, sonst mathematischer Fehler
    if (zd < (-1.0)) zd = -1.0;
    zd = 12.0 * acos(zd) / PI_KONST;                        // Zeitdifferenz

    sa = (12 - zd - zg) - geogr_l / 15.0 + ZEITZONE + zeit.mesz;    // Sonnenaufgang
    su = (12 + zd - zg) - geogr_l / 15.0 + ZEITZONE + zeit.mesz;    // Sonnenuntergang

    if (sa >= 24.0) sa = 0;                                 // Sicherheitsbegrenzungen
    if (sa < 0)     sa = 0;
    if (su >= 24.0) su = 0;
    if (su < 0)     su = 0;

    // Zeiten konvertieren und in Struktur bernehmen
    zeit.sauf_stunde = (unsigned char) sa;
    zeit.sauf_minute = (unsigned char) ((sa - zeit.sauf_stunde) * 60);
    zeit.sunt_stunde = (unsigned char) su;
    zeit.sunt_minute = (unsigned char) ((su - zeit.sunt_stunde) * 60);

    // Dmmerungszeiten berechnen
    switch (awd.astro_daemm) {
        case 0: // brgerliche Dmmerung
                h_const = H_KONST_6;
                break;
        case 1: // nautische Dmmerung
                h_const = H_KONST_12;
                break;
        case 2: // astronomische Dmmerung
                h_const = H_KONST_18;
    }
    zd = (sin(h_const) - sin(breite_gr) * sin(dk)) / (cos(breite_gr) * cos(dk));
    if (zd > 1.0)    zd =  1.0;                                 // Begrenzung, da ARCCOS ist nur von -1 bis +1 definiert, sonst mathematischer Fehler
    if (zd < (-1.0)) zd = -1.0;
    zd = 12.0 * acos(zd) / PI_KONST;                            // Zeitdifferenz

    sa = (12 - zd - zg) - geogr_l / 15.0 + ZEITZONE + zeit.mesz; // Morgendmmerung
    su = (12 + zd - zg) - geogr_l / 15.0 + ZEITZONE + zeit.mesz; // Abenddmmerung

    // Astronomische Dmmerung zwischen Tag 144 bis einschl. Tag 198 nicht nutzbar
    if (sa >= 24.0) sa = 0;    // Sicherheitsbegrenzungen, um SS-Wende ist Astromomische Dmmerung ungltig
    if (sa < 0)     sa = 0;
    if (su >= 24.0) su = 0;
    if (su < 0)     su = 0;

    // Zeiten konvertieren und in Struktur bernehmen
    zeit.md_stunde = (unsigned char) sa;
    zeit.md_minute = (unsigned char) ((sa - zeit.md_stunde) * 60);
    zeit.ad_stunde = (unsigned char) su;
    zeit.ad_minute = (unsigned char) ((su - zeit.ad_stunde) * 60);

    // Sonnenhchststand
    zeit.zg_minute = (char) (zg * 60) + 1;  // Abweichung der Mittagssonne in Minuten von 12:00 Uhr (mit Vorzeichen) 
    sonne_set      = 0;                     // Sonnenberechnung erledigt
  }

  // Ausschaltfunktion der Relais
  if (k1aus.zyklus && (k1aus.stunde == zeit.stunde) && (k1aus.minute == zeit.minute) && (k1aus.sekunde == zeit.sekunde) ) {
    // Relais K1 rcksetzen und LED aus
    k1aus.zyklus  = 0;
    k1aus.zustand = 0;
    K1_AUS;
    LED_D3_AUS;
  }   
  if (k2aus.zyklus && (k2aus.stunde == zeit.stunde) && (k2aus.minute == zeit.minute) && (k2aus.sekunde == zeit.sekunde) ) {
    // Relais K2 rcksetzen und LED aus
    k2aus.zyklus  = 0;
    k2aus.zustand = 0;
    K2_AUS;
    LED_D4_AUS;
  }   
  // Astro-Schaltprogramm
  if (!zeit.sekunde) {
    // immer zur vollen Minute auf Schaltaktivitt prfen
    for(s_index = 0 ; s_index < SPR_MAX ; s_index++) { 
        // alle Eintrge durchsuchen
        if (!spr[s_index].belegt) continue; // Eintrag nicht benutzt
        zeit_index  = zeit.stunde * 60 + zeit.minute;
        // Einschaltzeit berechnen
        switch (spr[s_index].e_fkt) {
            case 0: // SA Sonnenaufgang
                    zeit_start_index = zeit.sauf_stunde * 60 + zeit.sauf_minute;
                    break;
            case 1: // MD Morgendmmerung
                    zeit_start_index = zeit.md_stunde * 60 + zeit.md_minute;
                    break;
            case 2: // SU Sonnenuntergang
                    zeit_start_index = zeit.sunt_stunde * 60 + zeit.sunt_minute;
                    break;
            case 3: // AD Abenddmmerung
                    zeit_start_index = zeit.ad_stunde * 60 + zeit.ad_minute;
                    break;
            case 4: // FZ Festzeit
                    zeit_start_index = spr[s_index].e_stunde * 60 + spr[s_index].e_minute;
                    break;
        }
        // bei SA MD SU AD Offset dazurechnen
        if (spr[s_index].e_fkt < 4) {
            if (spr[s_index].e_vorz) {
                // negativ
                zeit_start_index -= (spr[s_index].e_stunde * 60 + spr[s_index].e_minute);
                if (zeit_start_index < 0) zeit_start_index = 1440 + zeit_start_index; // zeit_start_index ist negativ!!! bertrag zurck ber 00:00 Uhr
            } else {
                // positiv
                zeit_start_index += spr[s_index].e_stunde * 60 + spr[s_index].e_minute;
                if (zeit_start_index >= 1440) zeit_start_index -= 1440; // bertrag 24:00 Uhr
            }
        }
        // Ausschaltzeit berechnen
        switch (spr[s_index].a_fkt) {
            case 0: // SA Sonnenaufgang
                    zeit_stopp_index = zeit.sauf_stunde * 60 + zeit.sauf_minute;
                    break;
            case 1: // MD Abdendmmerung
                    zeit_stopp_index = zeit.md_stunde * 60 + zeit.md_minute;
                    break;
            case 2: // SU Sonnenuntergang
                    zeit_stopp_index = zeit.sunt_stunde * 60 + zeit.sunt_minute;
                    break;
            case 3: // AD Abenddmmerung
                    zeit_stopp_index = zeit.ad_stunde * 60 + zeit.ad_minute;
                    break;
            case 4: // FZ Festzeit
                    zeit_stopp_index = spr[s_index].a_stunde * 60 + spr[s_index].a_minute;
                    break;
        }
        // bei SA MD SU AD Offset dazurechnen
        if (spr[s_index].a_fkt < 4) {
            if (spr[s_index].a_vorz) {
                // negativ
                zeit_stopp_index -= (spr[s_index].a_stunde * 60 + spr[s_index].a_minute);
                if (zeit_stopp_index < 0) zeit_stopp_index = 1440 + zeit_stopp_index; // zeit_stopp_index ist negativ!!! bertrag zurck ber 00:00 Uhr
            } else {
                // positiv
                zeit_stopp_index += spr[s_index].a_stunde * 60 + spr[s_index].a_minute;
                if (zeit_stopp_index >= 1440) zeit_stopp_index -= 1440; // bertrag 24:00 Uhr
            }
        }
        // Schaltzeit mit Istzeit vergleichen
        if (zeit_index == zeit_start_index) {
            // Uhrzeit stimmt
            // Wochentag prfen Bit7=xx Bit6=Sa Bit5=Fr Bit4=Do Bit3=Mi Bit2=Di Bit1=Mo Bit0=So
            switch (zeit.wo_tag) {
                case 1: // Montag
                        if (spr[s_index].wo_tag & 0x02) ereignis_flag = 1;
                        break;
                case 2: // Dienstag
                        if (spr[s_index].wo_tag & 0x04) ereignis_flag = 1;
                        break;
                case 3: // Mittoch
                        if (spr[s_index].wo_tag & 0x08) ereignis_flag = 1;
                        break;
                case 4: // Donnerstag
                        if (spr[s_index].wo_tag & 0x10) ereignis_flag = 1;
                        break;
                case 5: // Freitag
                        if (spr[s_index].wo_tag & 0x20) ereignis_flag = 1;
                        break;
                case 6: // Samstag
                        if (spr[s_index].wo_tag & 0x40) ereignis_flag = 1;
                        break;
                case 7: // Sonntag
                        if (spr[s_index].wo_tag & 0x01) ereignis_flag = 1;
            }
            // Einschaltzeit und Wochentag stimmen berein
            if (ereignis_flag) {
                if (spr[s_index].k1) {
                    // Relais K1 einschalten / LED setzen
                    k1aus.stunde  = (unsigned char) (zeit_stopp_index / 60);
                    k1aus.minute  = (unsigned char) (zeit_stopp_index % 60);
                    k1aus.sekunde = 0;
                    k1aus.zyklus  = 1;
                    k1aus.zustand = 1;
                    K1_EIN;
                    LED_D3_EIN;
                 }
                if (spr[s_index].k2) {
                    // Relais K2 einschalten / LED setzen
                    k2aus.stunde  = (unsigned char) (zeit_stopp_index / 60);
                    k2aus.minute  = (unsigned char) (zeit_stopp_index % 60);
                    k2aus.sekunde = 0;
                    k2aus.zyklus  = 1;
                    k2aus.zustand = 1;
                    K2_EIN;
                    LED_D4_EIN;
                 }
            } // if ereigneis_flag
        } // if zeit_index
    } // for s_index
  } // if !zeit.sekunde

  LED_D1_EIN;                                               // LED D1 einschalten
  delay_counter = 10;                                       // Einschaltdauer 5=50ms 10=100ms
}

// ********************************************************************
// Timer 0 Interruptroutine
// Zeitinterval: 10ms
// ********************************************************************
ISR(TIMER0_OVF_vect) {
  TCNT0 = 148;                                              // 10ms bei 11.0592MHz Reloadwert=148
  if (!sys_status) {
    // Kein Fehler,LED D1 blinkt kurz (Aus-Sequenz)
     if (delay_counter) delay_counter--;    // Universalzhler 16-Bit
        else LED_D1_AUS;
  }
}
